/* pcl725io.c - Device driver for the Advantech PCL-725 and ADClone ACL-725
		relay actuator and isolated D/I cards */

/* Written 1996 by Werner Almesberger, EPFL LRC */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h> /* printk */
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include "pcl725.h"


#ifndef MODULE
#error "PCL-725/ACL-725 driver must be compiled as a module"
#endif


static unsigned short pcl725_map[PCL725_MAX_DEVS];
static unsigned char pcl725_mask[PCL725_MAX_DEVS];


static int pcl725_ioctl(struct inode *inode,struct file *filp,unsigned int cmd,
  unsigned long arg)
{
	unsigned short base;
	unsigned char mask;

	base = pcl725_map[MINOR(inode->i_rdev)];
	mask = pcl725_mask[MINOR(inode->i_rdev)];
	switch (cmd) {
		case PCL725_INP_GET:
			return inb_p(base+1);
		case PCL725_REL_GET:
			return inb_p(base) & mask;
		case PCL725_REL_SET:
			outb_p((inb_p(base) & ~mask) | (arg & mask),base);
			return 0;
		case PCL725_REL_ON:
			outb_p(inb_p(base) | (arg & mask),base);
			return 0;
		case PCL725_REL_OFF:
			outb_p(inb_p(base) & (~arg | ~mask),base);
			return 0;
		case PCL725_REL_TOGGLE:
			outb_p(inb_p(base) ^ (arg & mask),base);
			return 0;
		default:
			return -EINVAL;
	}
}


static int pcl725_open(struct inode *inode,struct file *filp)
{
	if (MINOR(inode->i_rdev) >= PCL725_MAX_DEVS) return -EINVAL;
	if (!pcl725_map[MINOR(inode->i_rdev)]) return -ENODEV;
	MOD_INC_USE_COUNT;
	return 0;
}


static void pcl725_release(struct inode *inode,struct file *filp)
{
	MOD_DEC_USE_COUNT;
}


static struct file_operations pcl725_fops = {
	NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	pcl725_ioctl,	/* ioctl */
	NULL,		/* mmap */
	pcl725_open,	/* open */
	pcl725_release,	/* release */
	NULL		/* fsync */
};


static int io[PCL725_MAX_DEVS*2];
static int major = PCL725_MAJOR;


int init_module(void)
{
	int i,j,error,index,needs_region;
	unsigned short base;
	unsigned char mask;

	if (register_chrdev(major,DEV_NAME,&pcl725_fops)) {
		printk("unable to get major %d for PCL-725/ACL-725 devs\n",
		    major);
		return -EIO;
	}
	memset(pcl725_map,0,sizeof(pcl725_map));
	error = 0;
	for (i = 0; i < PCL725_ADDRS; i++)
		if (io[i]) break;
	if (i == PCL725_ADDRS) io[0] = PCL725_DFL_ADDR;
	index = 0;
	for (i = 0; i < PCL725_MAX_DEVS*2; i++) 
		if (io[i]) {
			base = io[i];
			if (i+1 < PCL725_MAX_DEVS*2 && io[i+1] && io[i+1] <
			    0xff) mask = io[++i];
			else mask = 0xff;
			if (base < PCL725_BASE || base >= PCL725_BASE+
			    (PCL725_ADDRS << 1) || (base & 1)) {
				printk(DEV_NAME ": invalid address 0x%x\n",
				    base);
				error = -EINVAL;
				break;
			}
			if (index == PCL725_MAX_DEVS) {
				printk(DEV_NAME ": too many devices. "
				    "Increase PCL725_MAX_DEVS and recompile."
				    "\n");
				error = -E2BIG;
				break;
			}
			needs_region = 1;
			for (j = 0; j < index; j++)
				if (base == pcl725_map[j]) {
					needs_region = 0;
					break;
				}
			if (needs_region) {
				error = check_region(base,PCL725_PORTS);
				if (error) {
					printk(DEV_NAME ": region 0x%04x-"
					    "0x%04x is already in use\n",
					    base,base+PCL725_PORTS-1);
					break;
				}
			}
			outb_p(0,base);
			if (inb_p(base)) {
				printk(DEV_NAME ": no device at 0x%x\n",base);
				error = -ENODEV;
				break;
			}
			if (needs_region)
				request_region(base,PCL725_PORTS,DEV_NAME);
			printk(DEV_NAME "%d: registered at address 0x%x,"
			    " mask 0x%02x\n",index,base,mask);
			pcl725_map[index] = base;
			pcl725_mask[index++] = mask;
		}
	if (!error) return 0;
	while (--index >= 0) release_region(pcl725_map[index],PCL725_PORTS);
	(void) unregister_chrdev(major,DEV_NAME);
	return error;
}


void cleanup_module(void)
{
	int i;

	for (i = 0; i < PCL725_MAX_DEVS; i++)
		if (pcl725_map[i]) {
			outb_p(0,pcl725_map[i]);
			release_region(pcl725_map[i],PCL725_PORTS);
		}
	if (unregister_chrdev(major,DEV_NAME))
		printk("unable to unregister major %d for PCL-725/ACL-725 "
		    "devs\n",major);
}
